# -*- coding: utf-8 -*-

"""
ChEMBLFunctions.py - functions for interacting with ChEMBL Web Services

Use of this code is subject to the terms of the StarDrop License Agreement and
use, copying or redistribution of this file for use by other persons or
organisations is prohibited

Copyright(C) Optibrium Ltd 2019
stardrop-support@optibrium.com

ChEMBL Web Services was created and is maintained by EMBL-EBI
"""

import PythonQt
import ChEMBLQueryLogger
import requests
import json
import app
import socket
import time
import traceback

log = ChEMBLQueryLogger.ChEMBLQueryLogger('ChEMBLquerytool.log')
socket.setdefaulttimeout(60)

def _create_columns(column_definitions):
    """
    Create columns for a dataset based on a set of column definitions. This
    is a convenience function that allows us to cut down on lots of repeated
    code to make columns.

    :param column_definitions:  The list of columns to create. Each column
                                should consist of a name, a type and None or a
                                list of category names depending on type
    :return: A tuple containing column headers and category dictionaries
    """
    cols = list()

    for column_definition in column_definitions:
        header, col_type, categories = column_definition

        if col_type == app.mol:
            cols.append(app.Dataset.meta(header, app.mol))
        elif col_type == app.cont:
            cols.append(app.Dataset.createNumberHeader(header))
        elif col_type == app.string:
            cols.append(app.Dataset.meta(header))

    return cols


def chemblHyperLink(chemblid, link_type):
    """
        Create hyperlink for chembl IDs

        :param chemblid:  The ChEMBL ID
        :param link_type: The type of ID (assay or molecule)

        :return: A hyperlink string
    """

    if link_type == 'molecule':
        return ('<a href="https://www.ebi.ac.uk/chembl/compound/inspect/' +
                chemblid +
                '">' +
                chemblid +
                '</a>')

    if link_type == 'assay':
        return ('<a href="https://www.ebi.ac.uk/chembl/assay/inspect/' +
                chemblid +
                '">' +
                chemblid +
                '</a>')

def enter_target_chembl_ids(appui):
    """
        Function for requesting ChEMBL target ID and doing the API call

        :param: None

        :return activities: Data returned from Chembl Web services
        :return target_chembl_id: The ChEMBL ID(s) entered into the dialog box
    """

    dlg = PythonQt.IDQueryDialog()
    if dlg.run():
        # The return string is a target ID or multiple target IDs
        target_chembl_id = dlg.selectedQuery()
        log.log("running ID query %s" % (repr(target_chembl_id)))

        prog = PythonQt.SimpleProgress()
        prog.setWindowTitle("Finding activity data")
        prog.setMessage("Connecting to ChEMBL Web Services")
        prog.show()
        prog.setProgress(3)

        activities = []
        remove_me = ""

        # Limit the number of target ChEMBL IDs entered to 3 in order to limit the amount of data brought back into
        # StarDrop as large data sets will take a long time to process.
        if len(target_chembl_id) > 3:
            appui.show("Can't bring back data for more than three target ChEMBL IDs")
            dlg = PythonQt.IDQueryDialog()
            if dlg.run():
                # request the ChEMBL IDs again
                target_chembl_id = dlg.selectedQuery()
                log.log("running ID query %s" % (repr(target_chembl_id)))

        # If instead of being on separate lines, the IDs are separated by a comma, split these into the individual IDs.
        if ", " in target_chembl_id[0]:
            target_chembl_id = target_chembl_id[0].split(', ')
        # If the IDs are separated by a space, specify the format and ask again for IDs.
        elif " " in target_chembl_id[0]:
            appui.show("Please separate IDs with a comma or write one ID per line")
            dlg = PythonQt.IDQueryDialog()
            if dlg.run():
                # The return value is simply the list of IDs
                target_chembl_id = dlg.selectedQuery()
                log.log("running ID query %s" % (repr(target_chembl_id)))
                if ", " in target_chembl_id[0]:
                    target_chembl_id = target_chembl_id[0].split(', ')
        elif len(target_chembl_id) > 1 and " " in target_chembl_id[0]:
            appui.show("Please separate IDs with a comma or write one ID per line")
            dlg = PythonQt.IDQueryDialog()
            if dlg.run():
                # The return value is simply a tuple of IDs
                target_chembl_id = dlg.selectedQuery()
                log.log("running ID query %s" % (repr(target_chembl_id)))
                if ", " in target_chembl_id[0]:
                    target_chembl_id = target_chembl_id[0].split(', ')
        else:
            None

        # For each target ID, connect to ChEMBL Web Services (CWS) and return the data.
        # If the ID is entered is not capitalised, it will capitalise it and if it is entered as just the
        # number, it will add CHEMBL to it.

        increment = 10.0 / len(target_chembl_id)
        progress_start = 4

        # if the CHEMBL ID is lowercase, make it uppercase. If it is only the number add "CHEMBL"
        for target_id in target_chembl_id:
            original = target_id
            if target_id.isdigit():
                target_id = 'CHEMBL' + target_id
            else:
                None
            if target_id.isupper():
                None
            else:
                target_id = target_id.upper()

            prog.setProgress(progress_start)

            target_id = target_id.strip()

            result = target_API_call(target_id, prog, progress_start, increment)

            if not result:
                appui.show("Could not find any activities for " + original)
                remove_me = original
            elif result == "cancelled":
                None
            else:
                if activities:
                    for item in result:
                        activities.append(item)
                else:
                    activities = result

            progress_start += increment

        # If a target ChEMBL ID is not recognised, remove it from the list of IDs.
        target_chembl_id = tuple(x for x in target_chembl_id if x != remove_me)

        prog.hide()

        return activities, target_chembl_id

def target_API_call(target_id, prog, prog_int, increment):
    """
        Do the CWS API call

        One ID is given.
        Returns all binding activity data.

        :param target_id:    A ChEMBL target ID which the user has entered
        :param prog:         The progress bar
        :param prog_int:     Where the progress bar is when this function is called
        :param increment:    The increment which is determined by how many ChEMBL target IDs were entered
        :return results:     The binding activity results
        """

    prog.setProgress(prog_int)

    url = "https://www.ebi.ac.uk/chembl/api/data/activity?target_chembl_id=" + target_id + "&format=json&limit=1000"

    try:
        # timout set to 60 seconds to allow for larger returned data sets
        resolver = requests.get(url, timeout=60)
    except requests.HTTPError as err:
        if err.response.status_code == 400:
            warning_dlg = PythonQt.Warning()
            warning_dlg.setMessage("Error in input ID")
            warning_dlg.run()
            warning_dlg.delete()
    except Exception as e:
        print(e)
        warning_dlg = PythonQt.Warning()
        warning_dlg.setMessage("Connection error")
        warning_dlg.run()
        warning_dlg.delete()

    json_data = resolver.text

    result = None
    try:
        result = json.loads(json_data)
    except ValueError:
        pass

    # If there is more data then there will be a value for the 'next' key and this value is the url for the next
    # API call
    results = result.get('activities')
    next = result.get('page_meta').get('next')

    if int(result.get('page_meta').get('total_count')) > 8000:
        warning_dlg = PythonQt.Warning()
        warning_dlg.setMessage(
            "There are a large number of bioactivities for this target which will take a long time to load into StarDrop, would you like to continue?")
        if warning_dlg.run():
            # Get an increment value for the progress bar based on the size of the data set it is returning. The key
            # 'total_count' gives the total number of activity entries for that target ID.
            size_of_data = float((result.get('page_meta').get('total_count'))/1000)
            if size_of_data == 0:
                iteration = increment / 1.0
            else:
                iteration = increment / size_of_data

            # Until all data is returned, keep doing the API calls. Combine all the data together as the result.
            while next:
                if prog.visible:
                    prog_int += iteration
                    prog.setProgress(prog_int)
                    result_item, next = target_next_API_call(next)
                    for item in result_item:
                        results.append(item)
                else:
                    break

            return results
        else:
            return "cancelled"
    else:
        # Get an increment value for the progress bar based on the size of the data set it is returning. The key
        # 'total_count' gives the total number of activity entries for that target ID.
        size_of_data = float((result.get('page_meta').get('total_count')) / 1000)
        if size_of_data == 0:
            iteration = increment / 1.0
        else:
            iteration = increment / size_of_data

        # Until all data is returned, keep doing the API calls. Combine all the data together as the result.
        while next:
            if prog.visible:
                prog_int += iteration
                prog.setProgress(prog_int)
                result_item, next = target_next_API_call(next)
                for item in result_item:
                    results.append(item)
            else:
                break

        return results

def target_next_API_call(next):
    """
        Do API calls until there is no 'next'.

        Next is the url to use to bring back the next set of data.
        Returns a chunk of the binding activity data for that chunk as well as this next url.

        :param next:            URL to get the next chunk of data
        :return: activities:    Activity data
        :return: next:          The next next
        """

    url = "https://www.ebi.ac.uk" + next

    try:
        # timout set to 60 seconds to allow for larger returned data sets
        resolver = requests.get(url, timeout=60)
    except:
        warning_dlg = PythonQt.Warning()
        warning_dlg.setMessage("Connection error")
        warning_dlg.run()
        warning_dlg.delete()

    json_data = resolver.text

    result = None
    try:
        result = json.loads(json_data)
    except ValueError:
        pass

    next = result.get('page_meta').get('next')
    activities = result.get('activities')

    return activities, next

def return_activity_data_from_target(appui, activities, B_or_F):
    """
    A function for taking the JSON object returned from CWS and packaging it into a StarDrop data set.

    :param activities:    The data returned from CWS
    :param B_or_F:        Whether to return binding assay data (B) or the other data (F)
    :return: dataset:     A data set containing the activity information
    """
    try:
        prog = PythonQt.SimpleProgress()
        prog.setWindowTitle("Returning data")
        prog.setMessage("Processing...")
        prog.show()

        prog.setProgress(16)

        dict_of_smiles = {}
        list_of_basic_info = []

        # Create columns for the standard information and any activity data specific to the target
        column_definitions = [
            ("SMILES", app.mol, None),
            ("ChEMBL Molecule ID", app.string, None),
            ("Target Name", app.string, None),
            ("Target Organism", app.string, None)
        ]

        for activity in activities:
            if activity.get('assay_type') in B_or_F:
                if activity.get('value') == None:
                    None
                elif 'B' in B_or_F:
                    pchembl = "pChEMBL value (derived from type " + activity.get("type") + ")"
                    if (pchembl, app.cont, None) in column_definitions:
                        None
                    else:
                        column_definitions.append((pchembl, app.cont, None))
                else:
                    None

        prog.setProgress(17)

        for activity in activities:
            if activity.get('assay_type') in B_or_F:
                if activity.get('units') in (None, '-'):
                    header = activity.get('type')
                    if (header, app.cont, None) in column_definitions:
                        None
                    else:
                        column_definitions.append((header, app.cont, None))
                        column_definitions.append(('Assay ID for ' + header, app.string, None))
                else:
                    header = activity.get('type') + ' (' + activity.get('units') + ')'
                    if (header, app.cont, None) in column_definitions:
                        None
                    else:
                        column_definitions.append((header, app.cont, None))
                        column_definitions.append(('Assay ID for ' + header, app.string, None))

                if activity.get('canonical_smiles') in dict_of_smiles:
                    molecule_ID = dict_of_smiles.get(activity.get('canonical_smiles')).get(
                        'molecule_chembl_id')
                    if activity.get('molecule_chembl_id') == molecule_ID:
                        None
                    else:
                        if activity.get('value') == None:
                            None
                        else:
                            dict_of_smiles[activity.get('canonical_smiles')] = {
                                'molecule_chembl_id': activity.get('molecule_chembl_id')}
                            list_of_basic_info.append(
                                {'canonical_smiles': activity.get('canonical_smiles'),
                                 'molecule_chembl_id': activity.get(
                                     'molecule_chembl_id'), 'target_pref_name': activity.get(
                                    "target_pref_name"), 'target_organism': activity.get("target_organism")})

                else:
                    if activity.get('value') == None:
                        None
                    else:
                        dict_of_smiles[activity.get('canonical_smiles')] = {
                            'molecule_chembl_id': activity.get('molecule_chembl_id')}
                        list_of_basic_info.append(
                            {'canonical_smiles': activity.get('canonical_smiles'),
                             'molecule_chembl_id': activity.get(
                                 'molecule_chembl_id'), 'target_pref_name': activity.get(
                                "target_pref_name"), 'target_organism': activity.get("target_organism")})

        prog.setProgress(18)
        cols = _create_columns(column_definitions)
        dataset = app.Dataset(cols)

        if activities == None or not dict_of_smiles:
            appui.show("Could not find any activities")
            return None

        else:
            increment = 70.0 / len(list_of_basic_info)
            progress_start = 20

            # Package the activity data into a StarDrop data set
            for dict_of_info in list_of_basic_info:
                progress_start += increment
                prog.setProgress(progress_start)
                if prog.visible:
                    try:
                        dataset.cols[0].append(app.Dataset.molecule(dict_of_info.get("canonical_smiles")))
                    except:
                        dataset.cols[0].append(app.Dataset.molecule(""))
                    try:
                        mol_chembl_id = chemblHyperLink(dict_of_info.get("molecule_chembl_id"), 'molecule')
                        dataset.cols[1].append(
                            app.Dataset.string(mol_chembl_id))
                    except:
                        dataset.cols[1].append(
                            app.Dataset.string(""))
                    try:
                        dataset.cols[2].append(
                            app.Dataset.string(dict_of_info.get("target_pref_name")))
                    except:
                        dataset.cols[2].append(
                            app.Dataset.string(""))
                    try:
                        dataset.cols[3].append(
                            app.Dataset.string(dict_of_info.get("target_organism")))
                    except:
                        dataset.cols[3].append(
                            app.Dataset.string(""))
                    for i in range(len(column_definitions)):
                        list_of_multiples = []
                        list_of_assay_ids = []
                        list_of_pchembls = []
                        dict_of_qualifiers = {}
                        type_and_units = cols[i][3][0]
                        if i < 4:
                            field_added = True
                        else:
                            field_added = False
                            for activity in activities:
                                if activity.get('canonical_smiles') == dict_of_info.get("canonical_smiles"):
                                    if activity.get('assay_type') in B_or_F:
                                        if "pChEMBL value" in type_and_units:
                                            field_added = True
                                            if activity.get('assay_type') in B_or_F:
                                                header = "pChEMBL value (derived from type " + activity.get("type") + ")"
                                                if type_and_units == header:
                                                    pchembl = activity.get('pchembl_value')
                                                    if pchembl:
                                                        pchembl = float(pchembl)
                                                        list_of_pchembls.append(pchembl)
                                                    else:
                                                        None
                                        else:
                                            if activity.get('value') == "None":
                                                None
                                            else:
                                                if type_and_units.startswith("Assay ID"):
                                                    field_added = True
                                                else:
                                                    if activity.get('units') in (None, '', '-'):
                                                        header = activity.get('type')
                                                        if type_and_units == header:
                                                            try:
                                                                if activity.get('activity_comment') in (None, 'active', 'Active') or activity.get('activity_comment').isdigit():
                                                                    if activity.get('relation') in ('=', None):
                                                                        None
                                                                    else:
                                                                        dict_of_qualifiers[
                                                                            activity.get('value')] = activity.get(
                                                                            'relation')
                                                                    list_of_multiples.append(activity.get("value"))
                                                                    field_added = True
                                                                    list_of_assay_ids.append(
                                                                        activity.get("assay_chembl_id"))
                                                                else:
                                                                    None
                                                            except Exception as e:
                                                                traceback.print_exc()
                                                                print(e)

                                                    else:
                                                        header = activity.get('type') + ' (' + activity.get(
                                                            'units') + ')'
                                                        if type_and_units == header:
                                                            try:
                                                                if activity.get('activity_comment') in (
                                                                None, 'active', 'Active') or activity.get('activity_comment').isdigit():
                                                                    if activity.get('relation') in ('=', None):
                                                                        None
                                                                    else:
                                                                        dict_of_qualifiers[
                                                                            activity.get('value')] = activity.get(
                                                                            'relation')
                                                                    list_of_multiples.append(activity.get("value"))
                                                                    field_added = True
                                                                    list_of_assay_ids.append(
                                                                        activity.get("assay_chembl_id"))
                                                                else:
                                                                    None
                                                            except Exception as e:
                                                                traceback.print_exc()
                                                                print(e)
                                                                field_added = False

                        if "pChEMBL value" in type_and_units:
                            if list_of_pchembls:
                                if len(list_of_pchembls) == 1:
                                    val = float(list_of_pchembls[0])
                                    dataset.cols[i].append(
                                        app.Dataset.cont(val, 0.0))
                                elif len(list_of_pchembls) > 1:
                                    try:
                                        average = sum(map(float, list_of_pchembls)) / len(list_of_pchembls)
                                        average = float(average)
                                    except Exception as e:
                                        traceback.print_exc()
                                        print(e)
                                    dataset.cols[i].append(
                                        app.Dataset.cont(average, 0.0))
                                else:
                                    dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                            else:
                                dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))

                        assay_id_index = i + 1
                        # create a blank entry where there is no activity value for that compound
                        if not field_added:
                            dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                            dataset.cols[assay_id_index].append(app.Dataset.string(""))
                        else:
                            if list_of_multiples:
                                list_of_multiples = [_f for _f in list_of_multiples if _f]
                                list_of_assay_ids = [_f for _f in list_of_assay_ids if _f]
                                if len(list_of_multiples) > 1:
                                    if dict_of_qualifiers:
                                        # If there are multiple qualifiers and multiple values, average the multiple values so
                                        # long as all values have qualifiers and all the qualifiers are the same. If there are
                                        # different qualifiers or it is a mixture of values with and without qualifiers,
                                        # insert a blank.
                                        if len(dict_of_qualifiers) == len(list_of_multiples):
                                            qualifiers_list = []
                                            for key, value in list(dict_of_qualifiers.items()):
                                                if value in qualifiers_list:
                                                    None
                                                else:
                                                    qualifiers_list.append(value)
                                            if len(qualifiers_list) == 1:
                                                average = sum(map(float, list_of_multiples)) / len(list_of_multiples)
                                                average = float(average)
                                                qual = 0
                                                if qualifiers_list[0] == '<':
                                                    qual = 1
                                                elif qualifiers_list[0] == '>':
                                                    qual = 2
                                                dataset.cols[i].append((12 + 2 * 8 + 1, (True, [], []), (qual, average, 0.0)))
                                            else:
                                                dataset.cols[i].append(
                                                    (12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                        else:
                                            dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                    else:
                                        try:
                                            average = sum(map(float, list_of_multiples)) / len(list_of_multiples)
                                            average = float(average)
                                        except Exception as e:
                                            traceback.print_exc()
                                            print(e)
                                        dataset.cols[i].append(app.Dataset.cont(average, 0.0))
                                elif len(list_of_multiples) == 1:
                                    if dict_of_qualifiers:
                                        qual = 0
                                        if dict_of_qualifiers.get(list_of_multiples[0]) == '<':
                                            qual = 1
                                        elif dict_of_qualifiers.get(list_of_multiples[0]) == '>':
                                            qual = 2
                                        value = float(list_of_multiples[0])
                                        dataset.cols[i].append((12 + 2 * 8 + 1, (True, [], []), (qual, value, 0.0)))
                                    else:
                                        val = float(list_of_multiples[0])
                                        dataset.cols[i].append(app.Dataset.cont(val, 0.0))
                                else:
                                    dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                if list_of_assay_ids:
                                    if len(set(list_of_assay_ids)) == 1:
                                        assay_chembl_id = chemblHyperLink(list_of_assay_ids[0], 'assay')
                                        dataset.cols[assay_id_index].append(app.Dataset.string(assay_chembl_id))
                                    elif len(set(list_of_assay_ids)) > 1:
                                        list_of_hyperlinked = []
                                        for item in list_of_assay_ids:
                                            list_of_hyperlinked.append(chemblHyperLink(item, 'assay'))
                                        assay_id_string = ", ".join(list_of_hyperlinked)
                                        dataset.cols[assay_id_index].append(app.Dataset.string(assay_id_string))
                                    else:
                                        dataset.cols[assay_id_index].append(app.Dataset.string(""))

            list_of_cols_to_del = []

            # identify any empty columns, put them into a list
            for i in range(len(column_definitions)):
                valid_entry = False
                valid_pCHEMBL = False
                if i < 4:
                    valid_entry = True
                    valid_pCHEMBL = True
                else:
                    for field in dataset.cols[i]:
                        type_and_units = cols[i][3][0]
                        if 'pChEMBL value' in type_and_units:
                            valid_entry = True
                            if field[2][1] == 0:
                                None
                            else:
                                valid_pCHEMBL = True
                        else:
                            valid_pCHEMBL = True
                            if isinstance(field[2], str):
                                valid_entry = True
                            else:
                                if field[2][1] == 0:
                                    None
                                else:
                                    valid_entry = True

                if valid_entry:
                    None
                else:
                    list_of_cols_to_del.append(i)
                    list_of_cols_to_del.append(i + 1)
                if valid_pCHEMBL:
                    None
                else:
                    list_of_cols_to_del.append(i)

            prog.setProgress(95)

            # reverse the list of empty columns to delete them without affecting indexing
            reverse_list = list_of_cols_to_del[::-1]

            # delete the empty columns
            if reverse_list:
                for col_num in reverse_list:
                    dataset.delColumn(col_num)
            else:
                None

            if dataset.cols[0] and len(dataset.cols) < 5:
                appui.show("Compounds found but no activities associated with them, they may have been labelled inactive for this target")

            prog.hide()

    except Exception as e:
        traceback.print_exc()
        log.log(repr(e))

    return dataset

def selected_ids_query_activities(appui):
    """
    Perform an ID query.

    One or multiple IDs can be given.
    Creates a data set of all binding activity data, ChEMBL molecule IDs, target name,
    target organism and pChEMBL values.

    :param appui:    The stardrop appui object
    :return: None
    """
    try:
        activities, target_chembl_id = enter_target_chembl_ids(appui)

        if (not activities) or len(target_chembl_id) == 0:
            return
            
        if len(target_chembl_id) > 1:
            target_chembl_id = ", ".join(target_chembl_id)

        else:
            target_chembl_id = target_chembl_id[0]

        if activities:
            dataset = return_activity_data_from_target(appui, activities, ('B',))

        else:
            print("No data returned")

        if dataset.cols[0]:
            try:
                packdata = dataset.pack()
            except Exception as e:
                traceback.print_exc()
                print("type error: " + str(e))
            appui.exportDS(str(target_chembl_id) + " binding activities", packdata)


    except Exception as e:
        traceback.print_exc()
        log.log(repr(e))

def selected_ids_query_activities_functional(appui):
    """
    Perform an ID query.

    One or multiple IDs can be given.
    Creates a data set of all activities that aren't binding activity data, ChEMBL molecule IDs, target name,
    target organism and pChEMBL values.

    :param appui:    The stardrop appui object
    :return: None
    """
    try:
        activities, target_chembl_id = enter_target_chembl_ids(appui)
        
        if (not activities) or len(target_chembl_id) == 0:
            return

        if len(target_chembl_id) > 1:
            target_chembl_id = ", ".join(target_chembl_id)
        else:
            target_chembl_id = target_chembl_id[0]

        if activities:
            dataset = return_activity_data_from_target(appui, activities, ('F', 'A', 'P', 'U'))
        else:
            print("No data returned")

        if dataset.cols[0]:
            try:
                packdata = dataset.pack()
            except Exception as e:
                print("type error: " + str(e))
                traceback.print_exc()
                
            appui.exportDS(str(target_chembl_id) + " functional and other activities", packdata)

    except Exception as e:
        traceback.print_exc()
        log.log(repr(e))

def targets_and_activities_similar_comps(appui):
    """
    Function for searching ChEMBL for similar compounds and returning binding activity data for these compounds.
    It also outputs ChEMBL molecule IDs, similarity values and pChEMBL values into the data table.

    :param appui:    The stardrop appui object
    :return: None
    """
    try:
        dataset = return_similar_comps_dataset(appui, ('B'))
        if dataset.cols[0]:
            try:
                packdata = dataset.pack()
            except Exception as e:
                print("type error: " + str(e))
                traceback.print_exc()
            appui.exportDS("Binding activities", packdata)

    except Exception as e:
        log.log(repr(e))
        traceback.print_exc()

def targets_and_activities_similar_comps_functional(appui):
    """
    Function for searching ChEMBL for similar compounds and returning everything but binding activity data for
    these compounds. It also outputs ChEMBL molecule IDs, similarity values and pChEMBL values into the data table.

    :param appui:    The stardrop appui object
    :return: None
    """
    try:
        dataset = return_similar_comps_dataset(appui, ('F', 'A', 'P', 'U'))

        if dataset.cols[0]:
            try:
                packdata = dataset.pack()
            except Exception as e:
                print("type error: " + str(e))
                traceback.print_exc()
                
            appui.exportDS("Functional and other activities", packdata)

    except Exception as e:
        log.log(repr(e))
        traceback.print_exc()

def find_threshold(appui):
    """
    Function for requesting a similarity threshold percentage from the user.

    :param appui:       The stardrop appui object
    :return threshold:  The similarity threshold input by the user
    """
    threshold = (appui.query('Similarity',
                             'Enter a similarity percentage:\n(Note: This must be 70% or higher. If the percentage you choose is low, you may\nsee errors because too many results are returned)',
                             '97%'))
    # If user hadn't deleted the % sign, delete this
    if threshold.endswith('%'):
        threshold = threshold[:-1]
    else:
        None
    try:
        threshold = float(threshold)
        if isinstance(threshold, float):
            # The threshold must be an integer, otherwise the API won't return results
            threshold = int(round(threshold))

            # The threshold needs to be 70 or above, otherwise the API won't return results.
            if threshold < 70 or threshold > 100:
                threshold = find_threshold(appui)

            return (threshold)
        else:
            appui.show("Error in similarity threshold percentage")
            return None
    except:
        if threshold == "":
            return None
        else:
            appui.show("Error in similarity threshold percentage")
            return None


def return_similar_comps_dataset(appui, B_or_F):
    """
    A function for searching the ChEMBL database for similar compounds, getting the activity data for those compounds
    and packaging it into a StarDrop data set.

    :param appui:         The stardrop appui object
    :param B_or_F:        Whether to return binding assay data or not
    :return: dataset:     A data set containing the activity information
    """
    try:
        dataset = []
        # get currently selected molecule
        SMILES_input = appui.getCurrentMolecules()

        # if no molecule is selected, tell user to select one and re-run
        if not SMILES_input:
            appui.show("Please select a molecule before searching for similar compounds")

        elif len(SMILES_input) > 1:
            appui.show("Please select exactly one molecule")

        else:
            SMILES_input = SMILES_input[0]
            # Ask the user for a similarity threshold.
            threshold = find_threshold(appui)
            if threshold == None:
                None
            else:
                threshold_value = str(threshold)

                prog = PythonQt.SimpleProgress()
                prog.setWindowTitle("Finding activity data")
                prog.setMessage("Connecting to ChEMBL Web Services...")
                prog.show()
                prog.setProgress(2)

                # If the SMILES string has a . in it, it means the molecule is disconnected and CWS doesn't
                # recognise these SMILES. Could in future question the user whether they wants to run with
                # only the larger molecule, assuming CWS still doesn't recognise it.
                if "." in SMILES_input:
                    appui.show("Error due to disconnected molecule")
                else:
                    SMILES_url = ""

                    try:
                        SMILES_url = requests.utils.quote(SMILES_input)
                    except Exception as e:
                        traceback.print_exc()
                        print(e)

                    url = "https://www.ebi.ac.uk/chembl/api/data/similarity/" + SMILES_url + "/" + threshold_value + "?format=json&limit=1000"

     
                    try:
                        # timeout set to 45 seconds to allow for larger returned data sets
                        resolver = requests.get(url, timeout=45)
                    except Exception as e:
                        traceback.print_exc()
                        appui.show(e)

                    json_data = resolver.text

                    result = None
                    try:
                        result = json.loads(json_data)
                    except ValueError:
                        pass

                    try:
                        molecules = result.get("molecules")
                    except:
                        molecules = ""

                    SMILES_and_id_list = []
                    list_of_results = []
                    prog.setProgress(3)

                    if molecules:
                        for molecule in molecules:
                            molecule_result = molecule.get('molecule_chembl_id')
                            SMILES_ = molecule.get("molecule_structures").get("canonical_smiles")
                            similarity = molecule.get('similarity')

                            # URL for second API call where I am getting the bioactivity data for the similar
                            # compounds found in ChEMBL.
                            url2 = "https://www.ebi.ac.uk/chembl/api/data/activity?molecule_chembl_id=" + molecule_result + "&format=json&limit=5000"
                            activities = []
                            
                            resolver2 = None
                            json_data2 = None
                            try:
                                # timout set to 45 seconds to allow for larger returned data sets
                                resolver2 = requests.get(url2, timeout=45)
                                json_data2 = resolver2.text
                            except Exception as e:
                                print(e)

                            prog.setProgress(4)

                            result2 = None
                            try:
                                result2 = json.loads(json_data2)
                            except ValueError:
                                pass

                            try:
                                activities = result2.get('activities')
                            except Exception as e:
                                print(e)

                            if activities:
                                for activity in activities:
                                    list_of_results.append(activity)

                            if result2:
                                SMILES_and_id_list.append({'canonical_smiles': SMILES_, 'molecule_chembl_id': molecule_result,
                                                           'similarity': similarity})
                            else:
                                None

                    else:
                        appui.show("No similar compounds found in ChEMBL")
                    if list_of_results:
                        progression = 5.0 / len(list_of_results)
                    else:
                        progression = 5.0
                    progress_start = 4

                    column_definitions = [
                        ("SMILES", app.mol, None),
                        ("ChEMBL Molecule ID", app.string, None),
                        ("Similarity (%)", app.cont, None)
                    ]
                    if list_of_results:
                        for activity in list_of_results:
                            progress_start += progression
                            prog.setProgress(progress_start)
                            if activity.get('value') == None:
                                None
                            elif B_or_F == 'B':
                                if activity.get('target_pref_name') == "Unchecked":
                                    pchembl = 'Undefined target pChEMBL value (target: '  + activity.get('target_chembl_id') + ", type: " + activity.get("type") + ")"
                                else:
                                    pchembl = activity.get('target_pref_name') + ' pChEMBL value (target: ' + activity.get('target_chembl_id') + ", type: " + activity.get("type") + ")"
                                if (pchembl, app.cont, None) in column_definitions:
                                    None
                                else:
                                    column_definitions.append((pchembl, app.cont, None))
                            else:
                                None

                        for activity in list_of_results:
                            progress_start += progression
                            prog.setProgress(progress_start)
                            if activity.get('units') in (None, '', '-'):
                                if activity.get('target_pref_name') == "Unchecked":
                                    header = 'Undefined target ' + activity.get('type') + " (target: " + activity.get('target_chembl_id') + ")"
                                else:
                                    header = activity.get('target_pref_name') + ' ' + activity.get('type') + " (target: " + activity.get('target_chembl_id') + ")"
                                if (header, app.cont, None) in column_definitions:
                                    None
                                else:
                                    column_definitions.append((header, app.cont, None))
                                    column_definitions.append(('Assay ID for ' + header, app.string, None))
                            else:
                                if activity.get('target_pref_name') == "Unchecked":
                                    header = 'Undefined target ' + activity.get('type') + ' (' + activity.get('units') + ', target: ' + activity.get('target_chembl_id') + ")"
                                else:
                                    header = activity.get('target_pref_name') + ' ' + activity.get('type') + ' (' + activity.get('units') + ', target: ' + activity.get('target_chembl_id') + ")"
                                if (header, app.cont, None) in column_definitions:
                                    None
                                else:
                                    column_definitions.append((header, app.cont, None))
                                    column_definitions.append(('Assay ID for ' + header, app.string, None))

                    if column_definitions == 3:
                        appui.show("No similar compounds found in ChEMBL")
                    else:
                        prog.setWindowTitle("Returning data")
                        prog.setMessage("Processing...")

                        cols = _create_columns(column_definitions)
                        dataset = app.Dataset(cols)
                        progression = 70.0 / len(SMILES_and_id_list)
                        progress_start = 15

                        for mol in SMILES_and_id_list:
                            progress_start += progression
                            prog.setProgress(progress_start)
                            if prog.visible:
                                smiles = mol.get('canonical_smiles')
                                molecule_chemblid = chemblHyperLink(mol.get('molecule_chembl_id'), 'molecule')
                                similarity = float(mol.get('similarity'))
                                try:
                                    dataset.cols[0].append(app.Dataset.molecule(smiles))
                                except:
                                    dataset.cols[0].append(app.Dataset.molecule(""))
                                try:
                                    dataset.cols[1].append(app.Dataset.string(molecule_chemblid))
                                except:
                                    dataset.cols[1].append(app.Dataset.string(""))
                                try:
                                    dataset.cols[2].append(app.Dataset.cont(similarity, 0.0))
                                except:
                                    dataset.cols[2].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                for i in range(len(column_definitions)):
                                    type_and_units = cols[i][3][0]
                                    list_of_multiples = []
                                    list_of_assay_ids = []
                                    list_of_pchembls = []
                                    dict_of_qualifiers = {}
                                    if i < 3:
                                        field_added = True
                                    else:
                                        field_added = False
                                        if list_of_results:
                                            for activity in list_of_results:
                                                if type_and_units.startswith("Assay ID"):
                                                    field_added = True
                                                else:
                                                    if activity.get('assay_type') in B_or_F:
                                                        if activity.get('canonical_smiles') == smiles:
                                                            if "pChEMBL value" in type_and_units:
                                                                if activity.get('target_pref_name') == 'Unchecked':
                                                                    header = 'Undefined target pChEMBL value (target: '  + activity.get('target_chembl_id') + ", type: " + activity.get("type") + ")"
                                                                else:
                                                                    header = activity.get('target_pref_name') + ' pChEMBL value (target: ' + activity.get('target_chembl_id') + ", type: " + activity.get("type") + ")"
                                                                if type_and_units == header:
                                                                    pchembl = activity.get('pchembl_value')
                                                                    if pchembl:
                                                                        pchembl = float(pchembl)
                                                                        list_of_pchembls.append(pchembl)
                                                                        field_added = True
                                                                    else:
                                                                        None
                                                            else:
                                                                if activity.get('value') == "None":
                                                                    None
                                                                else:
                                                                    if activity.get('units') in (None, '', '-'):
                                                                        if activity.get('target_pref_name') == 'Unchecked':
                                                                            header = 'Undefined target ' + activity.get(
                                                                                'type') + " (target: " + activity.get(
                                                                                'target_chembl_id') + ")"
                                                                        else:
                                                                            header = activity.get(
                                                                                'target_pref_name') + ' ' + activity.get(
                                                                                'type') + " (target: " + activity.get(
                                                                                'target_chembl_id') + ")"
                                                                        if type_and_units == header:
                                                                            try:
                                                                                if activity.get('activity_comment') in (
                                                                                None, 'active', 'Active') or activity.get('activity_comment').isdigit():
                                                                                    if activity.get('relation') in ('=', None):
                                                                                        None
                                                                                    else:
                                                                                        dict_of_qualifiers[
                                                                                            activity.get('value')] = activity.get(
                                                                                            'relation')
                                                                                    list_of_multiples.append(activity.get("value"))
                                                                                    field_added = True
                                                                                    list_of_assay_ids.append(
                                                                                        activity.get("assay_chembl_id"))
                                                                                else:
                                                                                    None
                                                                            except Exception as e:
                                                                                print(e)

                                                                    else:
                                                                        if activity.get('target_pref_name') == 'Unchecked':
                                                                            header = 'Undefined target ' + activity.get(
                                                                                'type') + ' (' + activity.get(
                                                                                'units') + ', target: ' + activity.get(
                                                                                'target_chembl_id') + ")"
                                                                        else:
                                                                            header = activity.get(
                                                                                'target_pref_name') + ' ' + activity.get(
                                                                                'type') + ' (' + activity.get(
                                                                                'units') + ', target: ' + activity.get(
                                                                                'target_chembl_id') + ")"
                                                                        if type_and_units == header:
                                                                            try:
                                                                                if activity.get('activity_comment') in (
                                                                                None, 'active', 'Active') or activity.get('activity_comment').isdigit():
                                                                                    if activity.get('relation') in ('=', None):
                                                                                        None
                                                                                    else:
                                                                                        dict_of_qualifiers[
                                                                                            activity.get('value')] = activity.get(
                                                                                            'relation')
                                                                                    list_of_multiples.append(activity.get("value"))
                                                                                    field_added = True
                                                                                    list_of_assay_ids.append(
                                                                                        activity.get("assay_chembl_id"))
                                                                                else:
                                                                                    None
                                                                            except Exception as e:
                                                                                print(e)
                                                                                field_added = False

                                        if "pChEMBL value" in type_and_units:
                                            if list_of_pchembls:
                                                if len(list_of_pchembls) == 1:
                                                    val = float(list_of_pchembls[0])
                                                    dataset.cols[i].append(
                                                        app.Dataset.cont(val, 0.0))
                                                elif len(list_of_pchembls) > 1:
                                                    try:
                                                        average = sum(map(float, list_of_pchembls)) / len(list_of_pchembls)
                                                        average = float(average)
                                                    except Exception as e:
                                                        print(e)
                                                    try:
                                                        dataset.cols[i].append(app.Dataset.cont(average, 0.0))
                                                    except:
                                                        dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                                else:
                                                    dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))


                                        assay_id_index = i + 1
                                        # create a blank entry where there is no activity value for that compound
                                        if not field_added:
                                            if "pChEMBL value" in type_and_units:
                                                dataset.cols[i].append(
                                                    (12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                            else:
                                                dataset.cols[i].append(
                                                    (12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                                dataset.cols[assay_id_index].append(app.Dataset.string(""))
                                        else:
                                            if list_of_multiples:
                                                list_of_multiples = [_f for _f in list_of_multiples if _f]
                                                list_of_assay_ids = [_f for _f in list_of_assay_ids if _f]
                                                if len(list_of_multiples) > 1:
                                                    if dict_of_qualifiers:
                                                        # If there are multiple qualifiers and multiple values, average the multiple values so
                                                        # long as all values have qualifiers and all the qualifiers are the same. If there are
                                                        # different qualifiers or it is a mixture of values with and without qualifiers,
                                                        # insert a blank.
                                                        if len(dict_of_qualifiers) == len(list_of_multiples):
                                                            qualifiers_list = []
                                                            for key, value in list(dict_of_qualifiers.items()):
                                                                if value in qualifiers_list:
                                                                    None
                                                                else:
                                                                    qualifiers_list.append(value)
                                                            if len(qualifiers_list) == 1:
                                                                average = sum(map(float, list_of_multiples)) / len(
                                                                    list_of_multiples)
                                                                average = float(average)
                                                                qual = 0
                                                                if qualifiers_list[0] == '<':
                                                                    qual = 1
                                                                elif qualifiers_list[0] == '>':
                                                                    qual = 2
                                                                dataset.cols[i].append(
                                                                    (12 + 2 * 8 + 1, (True, [], []), (qual, average, 0.0)))
                                                            else:
                                                                dataset.cols[i].append(
                                                                    (12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                                        else:
                                                            dataset.cols[i].append(
                                                                (12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                                    else:
                                                        try:
                                                            average = sum(map(float, list_of_multiples)) / len(list_of_multiples)
                                                            average = float(average)
                                                            dataset.cols[i].append(
                                                                app.Dataset.cont(average, 0.0))
                                                        except Exception as e:
                                                            print(e)
                                                            dataset.cols[i].append((12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))
                                                elif len(list_of_multiples) == 1:
                                                    if dict_of_qualifiers:
                                                        qual = 0
                                                        if dict_of_qualifiers.get(list_of_multiples[0]) == '<':
                                                            qual = 1
                                                        elif dict_of_qualifiers.get(list_of_multiples[0]) == '>':
                                                            qual = 2
                                                        value = float(list_of_multiples[0])
                                                        dataset.cols[i].append((12 + 2 * 8 + 1, (True, [], []), (qual, value, 0.0)))
                                                    else:
                                                        try:
                                                            dataset.cols[i].append(
                                                                app.Dataset.cont(float(list_of_multiples[0]), 0.0))
                                                        except Exception as e:
                                                            print(e)
                                                else:
                                                    dataset.cols[i].append(
                                                        (12 + 2 * 8 + 1, (False, [], []), (0, 0, 0.0)))

                                                if len(set(list_of_assay_ids)) == 1:
                                                    assay_chembl_id = chemblHyperLink(list_of_assay_ids[0],
                                                                                      'assay')
                                                    dataset.cols[assay_id_index].append(
                                                        app.Dataset.string(assay_chembl_id))
                                                elif len(set(list_of_assay_ids)) > 1:
                                                    list_of_hyperlinked = []
                                                    for item in list_of_assay_ids:
                                                        list_of_hyperlinked.append(
                                                            chemblHyperLink(item, 'assay'))
                                                    assay_id_string = ", ".join(list_of_hyperlinked)
                                                    dataset.cols[assay_id_index].append(
                                                        app.Dataset.string(assay_id_string))
                                                else:
                                                    dataset.cols[assay_id_index].append(
                                                        app.Dataset.string(""))

                list_of_cols_to_del = []
                prog.setProgress(88)

                # identify any empty columns, put them into a list
                for i in range(len(column_definitions)):
                    valid_entry = False
                    valid_pCHEMBL = False
                    if i < 3:
                        valid_entry = True
                        valid_pCHEMBL = True
                    else:
                        for field in dataset.cols[i]:
                            type_and_units = cols[i][3][0]
                            if 'pChEMBL value' in type_and_units:
                                valid_entry = True
                                if field[2][1] == 0:
                                    None
                                else:
                                    valid_pCHEMBL = True
                            else:
                                valid_pCHEMBL = True
                                if isinstance(field[2], str):
                                    valid_entry = True
                                else:
                                    if field[2][1] == 0:
                                        None
                                    else:
                                        valid_entry = True


                    if valid_entry:
                        None
                    else:
                        list_of_cols_to_del.append(i)
                        list_of_cols_to_del.append(i + 1)
                    if valid_pCHEMBL:
                        None
                    else:
                        list_of_cols_to_del.append(i)

                prog.setProgress(95)

                # reverse the list of empty columns to delete them without affecting indexing
                reverse_list = list_of_cols_to_del[::-1]

                # delete the empty columns
                if reverse_list:
                    for col_num in reverse_list:
                        dataset.delColumn(col_num)
                else:
                    None

                if len(dataset.cols) == 3:
                    if len(dataset.cols[0]) == 1:
                        appui.show("Compound found, but no bioactivities")
                    else:
                        appui.show("Compounds found, but no bioactivities")

                prog.hide()

    except Exception as e:
        traceback.print_exc()
        log.log(repr(e))

    return dataset

def search_for_molecule_chemblID(appui):
    """
    Search the database for the compounds of the current data set. If the compound is in ChEMBL, return its
    ChEMBL molecule ID.

    The IDs are inserted into the current data set to the right of the structure column.

    :param appui:    The stardrop appui object
    :return: None
    """
    try:
        # Import the current data set to get the structures
        dataset = appui.importDS("")
        ds = app.Dataset()
        ds.parse(dataset)

        col_name_present = False
        number_of_named_cols = 0

        # Find the structure column
        structureColumn = -1
        chemblColumn = -1
        for colIndex in range(ds.columnSz()):
            type = ds.columnType(colIndex)
            if structureColumn == -1 and type == app.mol:
                structureColumn = colIndex
            if ds.columnName(colIndex).startswith("ChEMBL Molecule ID"):
                col_name_present = True
                number_of_named_cols += 1
        if structureColumn == -1:
            appui.show("No structures available to use in search")
            return False

        prog = PythonQt.SimpleProgress()
        prog.setWindowTitle("Finding ChEMBL IDs")
        prog.setMessage("Searching ChEMBL...")
        prog.show()
        prog.setProgress(3)

        # Set an increment for the progress bar based on the number of rows
        number_of_rows = ds.rowSz()
        num_of_rows_int = int(number_of_rows)
        progress_start = 5
        progression = 80.0 / (num_of_rows_int)
        quit = False

        if num_of_rows_int > 200:
            warning_dlg = PythonQt.Warning()
            warning_dlg.setMessage(
                "Data sets of over 200 compounds can take a long time to process. Would you like to proceed?")
            if not warning_dlg.run():
                quit = True

        if not quit:
            newEntries = []
            for rowIndex in range(ds.rowSz()):
                progress_start += progression
                prog.setProgress(progress_start)
                if prog.visible:
                    SMILES = ds.stringval(structureColumn, rowIndex)

                    SMILES_url = ""

                    try:
                        SMILES_url = requests.utils.quote(SMILES)
                    except Exception as e:
                        print(e)

                    # Do an API call for each individual compound
                    url = "https://www.ebi.ac.uk/chembl/api/data/similarity/" + SMILES_url + "/100?format=json"

                    try:
                        resolver = requests.get(url, timeout=30)
                    except Exception as e:
                        print(e)

                    json_data = resolver.text

                    result = None
                    try:
                        result = json.loads(json_data)
                    except ValueError:
                        pass

                    try:
                        molecules = result.get("molecules")
                    except:
                        molecules = ""

                    # If the compound is found in ChEMBL append a hyperlink of its ID to a list of data to go into
                    # the dataset. If it isn't found append "Not in CHEMBL".
                    if molecules:
                        mol_IDs_string = ""
                        for molecule in molecules:
                            mol_chembl_id = chemblHyperLink(molecule.get('molecule_chembl_id'), 'molecule')
                            if mol_IDs_string:
                                mol_IDs_string += ', ' + mol_chembl_id
                            else:
                                mol_IDs_string = mol_chembl_id
                        newEntries.append(app.Dataset.string(mol_IDs_string))
                    else:
                        if "." in SMILES_url:
                            newEntries.append(app.Dataset.string("Error due to disconnected molecule"))
                        else:
                            # If it isn't found, try the original SMILES
                            url = "https://www.ebi.ac.uk/chembl/api/data/similarity/" + SMILES + "/100?format=json"

                            try:
                                resolver = requests.get(url, timeout=30)
                            except Exception as e:
                                print(e)

                            json_data = resolver.text

                            result = None
                            try:
                                result = json.loads(json_data)
                            except ValueError:
                                pass

                            try:
                                molecules = result.get("molecules")
                            except:
                                molecules = ""
                            if molecules:
                                mol_IDs_string = ""
                                for molecule in molecules:
                                    mol_chembl_id = chemblHyperLink(molecule.get('molecule_chembl_id'), 'molecule')
                                    if mol_IDs_string:
                                        mol_IDs_string += ', ' + mol_chembl_id
                                    else:
                                        mol_IDs_string = mol_chembl_id
                                newEntries.append(app.Dataset.string(mol_IDs_string))
                            else:
                                newEntries.append(app.Dataset.string("Not in ChEMBL"))

        else:
            None

        prog.setProgress(89)

        if chemblColumn == -1:
            chemblColumn = structureColumn + 1
        else:
            appui.removeColumnFromDataSet("", chemblColumn)

        # If there is already one or several columns called "ChEMBL Molecule ID", add <number> to the column name.
        if col_name_present:
            columnName = "ChEMBL Molecule ID <" + str(number_of_named_cols) + ">"
            columnHeader = app.Dataset.meta(columnName, app.string)
            ds.insertColumn(columnHeader, newEntries, chemblColumn)
            appui.addColumnToDataSet("", chemblColumn, ds.packColumn(chemblColumn))
        else:
            columnName = "ChEMBL Molecule ID"
            columnHeader = app.Dataset.meta(columnName, app.string)
            ds.insertColumn(columnHeader, newEntries, chemblColumn)
            appui.addColumnToDataSet("", chemblColumn, ds.packColumn(chemblColumn))

        prog.setProgress(95)
        prog.hide()

    except Exception as e:
        traceback.print_exc()
        log.log(repr(e))


def about(appui):
    """
    This is to provide general information as well as the current status of the web API.

    :param appui:
    :return: None
    """

    try:
        # Get the current status of the API and API information
        url = "https://www.ebi.ac.uk/chembl/api/data/status?format=json"

        try:
            resolver = requests.get(url, timeout=30)
        except Exception as e:
            print(e)

        try:
            json_data = resolver.text
        except Exception as e:
            print(e)

        result = None

        try:
            result = json.loads(json_data)
        except Exception as e:
            print(e)

        if result:
            status_report_text = "ChEMBL Web Services is currently " + result.get(
                "status") + ". The API version is " + result.get(
                "api_version") + ". The ChEMBL database version is " + result.get(
                "chembl_db_version") + ", released on " + result.get("chembl_release_date") + "."
        else:
            status_report_text = "Unable to connect to ChEMBL Web Services, please check internet connection."

        # Show general information and API status
        appui.show('ChEMBL Web Services<br>EMBL-EBI 2019 <br><br>ChEMBL is a publicly available database of drugs, small molecules and their targets. It contains information on more than 1.4 million compounds and over 12 million records of their effects on biological systems.<br>To find ChEMBL Target IDs, follow this link to the ChEMBL website: <a href="https://www.ebi.ac.uk/chembl">ChEMBL</a><br><br>Status Report:<br><br>' + status_report_text)

    except Exception as e:
        traceback.print_exc()
        log.log(repr(e))


def get_stardrop_definitions():
    definitions = [
        {
            'script_name': 'ChEMBL Web Services/Search by ChEMBL Target ID/Binding Assays...',
            'callback': selected_ids_query_activities
        },
        {
            'script_name': 'ChEMBL Web Services/Search by ChEMBL Target ID/Functional and other assays...',
            'callback': selected_ids_query_activities_functional
        },
        {
            'script_name': 'ChEMBL Web Services/Search by Similarity to Selected Compound/Binding assays...',
            'callback': targets_and_activities_similar_comps
        },
        {
            'script_name': 'ChEMBL Web Services/Search by Similarity to Selected Compound/Functional and other assays...',
            'callback': targets_and_activities_similar_comps_functional
        },
        {
            'script_name': 'ChEMBL Web Services/Add ChEMBL Molecule IDs to Data Set',
            'callback': search_for_molecule_chemblID
        },
        {
            'script_name': 'ChEMBL Web Services/About',
            'callback': about
        }
    ]

    return definitions